# -*- coding: utf-8 -*-
########################################################################
# 
# Copyright (c) 2019 Baidu.com, Inc. All Rights Reserved
# 
########################################################################
 

import sys
import threading
from collections import OrderedDict

from genregion.region import geometry as mygeo
from genregion.region import error

class Link(mygeo.Segment):
    """The link object designed for some region operations.

    Attributes:
        twin_link (link): The link at the same position but with opposite direction.
        region (region): The region should always be at the left side of the link.

    """
    #__slots__ = ("twin_link", "region")

    def __init__(self, start, end):
        """Initialize the link object.

        Args:
            start (point): The start point.
            end (point): The end point.
        """
        mygeo.Segment.__init__(self, start, end)
        self.twin_link = None
        self.region = None
    
    def twin_link_region(self):
        """Get the region of the twin link.

        Returns:
            region: The left region of the twin link.
        """
        ret = None
        if self.twin_link is not None:
            ret = self.twin_link.region
        return ret


class Neighbor(object):
    """The neighbor region object.

    Attributes:
        region (region): A region in the neighborhood.
        points (list): A list of points that represents the contiguous border of two regions.
        length (float): The length of the contiguous border of two regions.
    """
    #__slots__ = ("region", "points", "length")

    def __init__(self, region, points, length):
        """Initialize the neighbor object.

        Args:
            region (region): One region in the neighborhood.
            points (list): A list of points that represent the contiguous border of two regions.
            length (float): The length of the contiguous border of two regions.
        """
        self.region = region
        self.points = points
        self.length = length

    def merge(self, points, length):
        """Merge a list of points to this neighbor.

        Note that the merging process can only happen when the last point of the neighbor is the last
        point of the new list of points.

        Args:
            points (list): A list of points that waited to be merged.
            length (float): The original length of the contiguous border.

        Raises:
            error.RegionError: Two lists of points are not connected.
        """
        if not points[-1] == self.points[0]:
            raise error.RegionError("Two lists of points are not connected.")
        newpoints = []
        newpoints.extend(points)
        newpoints.extend(self.points[1:])
        self.points = newpoints
        self.length += length

    def display(self):
        """Print the information of this neighbor.
        """
        print("neighbor")
        if self.region is not None:
            print("\tregion: ", str(self.region))
        else:
            print("\tregion: None")
        print("\tPoint: ")
        for pt in self.points:
            print(str(pt), ", ")
        print("\tlength: ", self.length)


class LinkDict(object):
    """The dictionary for links.

    Attributes:
        link_dict (dict): A dictionary contains all links.

    """
    #__slots__ = ("link_dict", )

    def __init__(self):
        """Initialize the LinkDict with an empty dictionary.
        """
        # self.link_dict = {}
        self.link_dict = OrderedDict()

    def clear(self):
        """Clear the dictionary.
        """
        # self.link_dict = {}
        self.link_dict = OrderedDict()

    def get_link(self, start, end):
        """Get or add a link using the start and the end point.

        Note that we first try to find the link in the dictionary and return it.
        If it is not existed in the dictionary, we add it to the dictionary and return it.

        Args:
            start (mygeo.Point): The start point.
            end (mygeo.Point): The end point.

        Returns:
            link: A directional link.
        """
        ret = None
        link = Link(start, end)
        linkstr = link.askey()
        if linkstr in self.link_dict:
            ret = self.link_dict[linkstr]
            if not start == ret.start:
                ret = ret.twin_link
        else:
            twin_link = Link(end, start)
            link.twin_link = twin_link
            twin_link.twin_link = link
            self.link_dict[linkstr] = link
            ret = link
        return ret

    def del_link(self, start, end):
        """Delete a link from the link dictionary.

        Args:
            start (mygeo.Point): The start point.
            end (mygeo.Point): The end point.
        """
        linkstr = Link(start, end).askey()
        if linkstr in self.link_dict:
            del self.link_dict[linkstr]

    def segments(self):
        """Get all undirectional segments from the link dictionary.

        Returns:
            list: A list of segments.
        """
        ret = []
        for link in self.link_dict.values():
            ret.append(mygeo.Segment(link.start, link.end))
        return ret


def points_2_segs(points):
    """Convert a list of points into a list of segments.

    Args:
        points (list): A list of mygeo.Point objects.

    Returns:
        list: A list of segments.
    """
    ret = []
    lastpt = points[0]
    for pt in points[1:]:
        ret.append(mygeo.Segment(lastpt, pt))
        lastpt = pt
    return ret


class Region(mygeo.Region):
    """The region object designed to facilitate some region operations.

    Attributes:
        points (list): A list of points that indicates the outer loop of the regions.
        holes (list): A list of lists of points that indicates holes.
        link_dict (LinkDict): A dictionary of links that stores all links we need.
        parent_region (region): The original region that contains this region.
        links (list): A list of lists of links.
        area (float): The area of this region.
        length_ (float): The perimeter of this region.
        width_ (float): A width that trying to indicate whether the region is too narrow.
        neighbors (list): A list of neighbors.
    """
    #__slots__ = ("link_dict", "parent_region", "links", "area_", "length_", "width_", \
                 #"neighbors")

    def __init__(self, region, parent_region, link_dict):
        """Initialize the region.

        Args:
            region (mygeo.Region): The mygeo.Region object used to initialize the region.
            parent_region (region): The region that originally contains this region.
            link_dict (LinkDict): A dictionary that stores all needed links.
        """
        mygeo.Region.__init__(self, region.points, region.holes)
        self.link_dict = link_dict
        self.parent_region = parent_region
        self.links = self.__get_links()
        self.area_ = region.area()
        self.length_ = region.length()
        self.width_ = self.area_ / self.length_
        self.neighbors = None

    def __trans_points_to_links(self, points):
        """Convert a list of points to a list of links.

        Args:
            points (list): A list of links.

        Returns:
            list: A list of links.
        """
        ret = []
        lastpt = points[-1]
        for pt in points:
            link = self.link_dict.get_link(lastpt, pt)
            link.region = self
            ret.append(link)
            lastpt = pt
        return ret

    def __get_links(self):
        """Convert a region in to links.

        Returns:
            list: A list of lists of links that represents the region and it's holes.
        """
        links = self.__trans_points_to_links(self.points)
        ret = [links]
        for hole in self.holes:
            links = self.__trans_points_to_links(hole)
            ret.append(links)
        return ret
        
    def __compute_ring_neighbors(self, links):
        """Get all neighbors of a list of links acquired from a ring.

        Args:
            links (list): A list of links acquired from a ring.

        Returns:
            list: A list of neighbors. 
        """
        ret = []
        if len(links) == 0:
            return ret
        points = []
        length = 0
        last_region = -1
        for link in links:
            linklen = link.length()
            region = link.twin_link_region()
            if region == last_region:
                points.append(link.end)
                length += linklen
            else:
                if last_region != -1:
                    ret.append(Neighbor(last_region, points, length))
                points = [link.start, link.end]
                length = linklen
                last_region = region
        if last_region != -1:
            if len(ret) > 1 and last_region == ret[0].region:
                ret[0].merge(points, length)
            else:
                ret.append(Neighbor(last_region, points, length))
        return ret
        
    def refresh_neighbors(self):
        """Get all neighbors of this region.
        """
        ret = []
        for links in self.links:
            neighbors = self.__compute_ring_neighbors(links)
            ret.extend(neighbors)
        self.neighbors = ret

    def is_small_region(self, area_thres, width_thres):
        """Check whether the current region is a small region.

        There are two types of small regions.
        The first type is the one that has a very small area.
        The second type is the one that is so narrow that can be considered as a road in reality.

        Args:
            area_thres (int): The threshold of a small area.
            width_thres (int): The threshold of a narrow width

        Returns:
            bool: True if it is a small region.
        """
        return self.area_ <= area_thres or self.width_ <= width_thres
            
    def merge_small_region(self, area_thres, width_thres):
        """Get all segments that need to be removed from this region.

        The main purpose of this function is to merge a small region to a larger one, 
        or merge a small region to another small region to get a larger one.
        A very simple way of merging two regions is to remove their contiguous border.
        So, we first try to find a non-small neighbor which has the largest contiguous border of his region.
        If the region does not have a non-small neighbor, we find the neighbor that has the largest contiguous border.

        Args:
            area_thres (int): The threshold of the area.
            width_thres (int): The threshold of the width.

        Returns:
            list: A list of segments that denotes the contiguous border.
        """
        small = True   # The current region is a small region.
        max_length = 0
        max_region = None
        for neighbor in self.neighbors:
            if neighbor.region is None:
                continue
            if not neighbor.region.is_small_region(area_thres, width_thres):
                if small or neighbor.length > max_length: # Find the region that has the largest length.
                    small = False
                    max_length = neighbor.length
                    max_region = neighbor.region
            elif small and neighbor.length > max_length: 
                max_length = neighbor.length
                max_region = neighbor.region
        if max_region is not None:
            for nb in self.neighbors:
                if nb.region == max_region:
                    return points_2_segs(nb.points)
        return None
    
    def merge_sibling_region(self):
        """Get all segments of all sibling regions of this region.

        Returns:
            list: A list of segmetns.
        """
        if self.parent_region is None:
            return None
        ret = []
        for nb in self.neighbors:
            if nb.region is not None and nb.region.parent_region == self.parent_region:
                ret.extend(points_2_segs(nb.points))
        return ret



def build_region_dict(regions, grid_size, key=None):
    """Construct the dictionary of grids and regions.

    The key is the grid, and the corresponding value is a list of regions that appears on that grid.

    Args:
        regions (list): A list of regions for us to construct the dictionary.
        grid_size (int): The size of the grid.
        key (tuple, optional): The integer coordinates of the grid. Defaults to None.

    Returns:
        dict: A dictionary that helps us find regions according to the grid information.
    """
    if key is None:
        key = lambda x: x
    grid_dict = OrderedDict()
    for reg in regions:
        for grid in key(reg).grids(grid_size):
            if grid in grid_dict:
                grid_dict[grid].append(reg)
            else:
                grid_dict[grid] = [reg]
    return grid_dict


class RegionMerger(object):
    """Merge small regions to larger ones.

    Attributes:
        regions (list): A list of regions that needed to be merged.
        raw_regions (list): A list of raw regions. However, we temporarily shut down this function.
        raw_region_dict (dict): A dictionary that stores all raw regions, and we do not need to use it.
        link_dict (LinkDict): A dictioanry which stores all links.

    """
    def __init__(self, regions, raw_regions):
        """Initialize the region merger.

        Args:
            regions (list): A list of regions waiting to be merged.
            raw_regions (list): We temporarily do not need to consider this.
        """
        self.regions = regions
        self.raw_regions = raw_regions
        self.raw_region_dict = None
        self.link_dict = LinkDict()

    def __get_raw_region(self, region, grid_size):
        """We temporarily do not need to consider this function.
        """
        if self.raw_region_dict is None:
            return None
        center = region.center()
        grid = center.grid(grid_size)
        if grid in self.raw_region_dict:
            raw_regions = self.raw_region_dict[grid]
            for rawreg, mbr in raw_regions:
                if mbr[1].x >= center.x and mbr[1].y >= center.y \
                   and mbr[0].x <= center.x and mbr[0].y <= center.y:
                    if rawreg.polygon().covers(center.point()):
                        return rawreg
        return None

    def __trans_regions(self, grid_size):
        """Transform the region type.

        Args:
            grid_size (int): The grid size.

        Returns:
            list: A list of transformed regions.
        """
        ret = []
        if self.raw_regions is not None:
            raw_regions = [(r, r.mbr()) for r in self.raw_regions]
            self.raw_region_dict = build_region_dict(raw_regions, grid_size, lambda x:x[0])
        for region in self.regions:
            raw_region = self.__get_raw_region(region, grid_size)
            newreg = Region(region, raw_region, self.link_dict)
            ret.append(newreg)
        for reg in ret:
            reg.refresh_neighbors()
        return ret

    def run(self, grid_size, area_thres, width_thres):
        """Merge regions based on given thresholds.

        Args:
            grid_size (int): The size of the grid.
            area_thres (int): Any region that has a smaller area than this will be merged.
            width_thres (int): Any region that has a smaller width than this will be merged.

        Returns:
            list: A list of segments that constructs those new regions.
        """
        regions = self.__trans_regions(grid_size) # mygeo.Region -> Region
        for region in regions:
            segs = region.merge_sibling_region()
            if segs is not None:
                for seg in segs:
                    self.link_dict.del_link(seg.start, seg.end)
            elif region.is_small_region(area_thres, width_thres):
                segs = region.merge_small_region(area_thres, width_thres)
                if segs is not None:
                    for seg in segs:
                        self.link_dict.del_link(seg.start, seg.end)
        return self.link_dict.segments()


class RegionFilter(object):
    """Filter out subregions among other regions.

    Attributes:
        regions (list): A list of regions.
    """
    def __init__(self, regions):
        """Initialize the filter.

        Args:
            regions (list): A list of regions that waited to be filtered.
        """
        self.regions = regions

    def __delete_contain(self, grid_dict):
        """Delete subregions.

        If more than half of the region is contained in the other region,
        we also consider it as the subregion.

        Args:
            grid_dict (dict): The grid dictionary that helps locate the region.
        """
        total = len(grid_dict)
        for dict_regions in grid_dict.values():
            regions = [r for r in dict_regions if not r[0].is_empty()]
            regions.sort(key=lambda x: x[1])
            for i in range(len(regions) - 1):
                child_reg, child_area, child_mbr = regions[i]
                for (parent_reg, parent_area, parent_mbr) in regions[i + 1:]:
                    if parent_mbr[1].x >= child_mbr[1].x and parent_mbr[1].y >= child_mbr[1].y \
                      and parent_mbr[0].x <= child_mbr[0].x and parent_mbr[0].y <= child_mbr[0].y:
                        try:
                            if (parent_reg.polygon().covers(child_reg.polygon())):
                                child_reg.destroy()
                        except error.RegionError:
                            continue
            
    def run(self, grid_size):
        """Filter the region.

        Args:
            grid_size (int): The size of the grid.

        Returns:
            list: A list of filtered regions.
        """
        # grid_dict = {}
        grid_dict = OrderedDict()
        for reg in self.regions:
            area = reg.area()
            mbr = reg.mbr()
            for grid in reg.grids(grid_size):
                if grid in grid_dict:
                    grid_dict[grid].append((reg, area, mbr))
                else:
                    grid_dict[grid] = [(reg, area, mbr)]
        self.__delete_contain(grid_dict)
        return [r for r in self.regions if not r.is_empty()]

    def __thread_start(self, func, grid_dict_list):
        """Start a thread for a function.

        Args:
            func (function): The function we want to run.
            grid_dict_list (list): A list of grids we want to input into the function.
        """
        thds = []
        for gd in grid_dict_list:
            t = threading.Thread(target=func, args=(gd,))
            t.start()
            thds.append(t)
        for t in thds:
            t.join()

    def multi_run(self, grid_size, thread_num=4):
        """Run the region filter through multiple threads.

        Args:
            grid_size (int): The size of the grid.
            thread_num (int, optional): Number of threads we want to use. Defaults to 4.

        Returns:
            list: A list of filtered regions.
        """
        grid_dict_list = []
        for i in range(thread_num):
            # grid_dict_list.append({})
            grid_dict_list.append(OrderedDict())
        for reg in self.regions:
            area = reg.area()
            mbr = reg.mbr()
            for grid in reg.grids(grid_size):
                idx = (grid[0] + grid[1]) % thread_num
                grid_dict = grid_dict_list[idx]
                if grid in grid_dict:
                    grid_dict[grid].append((reg, area, mbr))
                else:
                    grid_dict[grid] = [(reg, area, mbr)]
        self.__thread_start(self.__delete_contain, grid_dict_list)
        return [r for r in self.regions if not r.is_empty()]


class RegionFinder(object):
    """Find the id of the region among a list of regions.

    Attributes:
        grid_size (int): The size of the grid.
        region_dict (dict): The region dictioanry.
    """
    def __init__(self, regions, grid_size):
        """
        Desc: 初始化

        Args:
            self : self

            regions : list of (id, region)

            grid_size : 网格大小
        Return: 
            None 
        Raises: 
            None
        """
        self.grid_size = grid_size
        self.region_dict = build_region_dict(regions, grid_size, key=lambda x: x[1])

    def find_region(self, point):
        """Find the id of the region that the given point is located at.

        Args:
            point (point): The given point.

        Returns:
            optional: The id of the region is determined by the user.
        """
        grid = point.grid(self.grid_size)
        if grid not in self.region_dict:
            return None
        for id, region in self.region_dict[grid]:
            if region.polygon().covers(point.point()):
                return id
        return None





# -*- coding: utf-8 -*-
########################################################################
# 
# Copyright (c) 2019 Baidu.com, Inc. All Rights Reserved
# 
########################################################################
 

import math
from collections import OrderedDict

from genregion.region import error
from genregion.region import geometry as mygeo
from genregion.generate.gen import cluster

def cos(invec, outvec):
    """Calculate the cosine value for two vectors.

    For each node, an invec means the vector that points to it, whereas an outvec means the vector that starts from it.
    This fuction is mainly used to calculate the cosine value of the invec and outvec of a node.

    Args:
        invec (tuple): A tuple of coordinates that represents the vector.
        outvec (tuple): A tuple of coordinates that represents the vector.

    Raises:
        error.RegionError: Zero vector.

    Returns:
        float: The cosine value.
    """
    (inx, iny) = invec
    (outx, outy) = outvec
    if (inx == 0 and iny == 0) or (outx == 0 and outy == 0):
        raise error.RegionError("Zero vector.")
    inlen = math.sqrt(inx * inx + iny * iny)
    outlen = math.sqrt(outx * outx + outy * outy)
    return (1.0 * inx * outx + 1.0 * iny * outy) / (inlen * outlen)

    
def side(invec, outvec):
    """Calculate the relative position of the outvec with respect to the invec.

    Here we use the cross product of two vectors to calculate the pseudo-sine value.
    We call it 'pseudo' since we don't divide those two norms, because we only have to know whether it is greater than 0 or not.
    If this value is less than 0, we define the outvec is on the right side of the invec.
    If this value is larger than 0, we define the outvec is on the left side of the invec.
    However, if the invec and the outvec are having the same or the opposite directions, the pseudo-sine value will be 0 iindistinctively.
    So we need the cosine angle of those two vectors to come in and help.
    When the pseudo-sine value is 0, if the cosine value is greater than 0, then the invec and the outvec must have the same direction;s
    on the contrary, if the cosine value is less than 0, those two vectors must have the opposite directions.


    Args:
        invec (tuple): A tuple of coordinates that represents the vector.
        outvec (tuple): A tuple of coordinates that represents the vector.

    Raises:
        error.RegionError: Zero vector.

    Returns:
        int: 1 represents the outvec is on the right. -1 represents the other way.
    """
    (inx, iny) = invec
    (outx, outy) = outvec
    if (inx == 0 and iny == 0) or (outx == 0 and outy == 0):
        raise error.RegionError("Zero vector.")
    side1 = inx * outy - outx * iny
    cos1 = inx * outx + iny * outy
    if side1 < 0:   
        return -1
    elif side1 > 0:  
        return 1
    elif cos1 > 0:   
        return 1
    else:   
        return -1


class Node(mygeo.Point):
    """ The point object for region generation.

    The difference between the node and the mygeo.Point is that we record the invecs and the outvecs.

    Attributes:
        x (float): The x-coordinate.
        y (float): The y-coordinate.
        in_links (list): Store all links that end at this node.
        out_links (list): Store all links that start at this node.

    """
    def __init__(self, x, y):
        """Initialization of the Node.

        Args:
            x (float): The x-coordinate.
            y (float): The y-coordinate.
        """
        mygeo.Point.__init__(self, x, y)
        self.in_links = []
        self.out_links = []


class Link(mygeo.Segment):
    """The segment object for region generation.

    Attributes:    
        start (node): The starting node of a link.
        end (node): The ending node of a link.
        used (bool): Check whether this linke is used during the region generation process.

    """
    def __init__(self, start, end):
        """The segment consists of two nodes.

        Args:
            start (node): The starting node of a link.
            end (node): The ending node of a link.
        """
        mygeo.Segment.__init__(self, start, end)
        start.out_links.append(self)
        end.in_links.append(self)
        self.used = False
        
    def leftest(self):
        """Get the leftest outlink of the current link.

        Raises:
            error.RegionError: No leftest link.

        Returns:
            Link: The leftest Link.
        """
        min_cs = 5.0
        min_link = None
        for link in self.end.out_links:
            out_vec = link.vec()
            in_vec = self.vec()
            side1 = side(in_vec, out_vec)
            cos1 = cos(in_vec, out_vec)
            cs = side1 * cos1 + (1 - side1)
            if cs < min_cs:
                min_cs = cs
                min_link = link
        if min_link is None:
            raise error.RegionError("There is no leftest link.")
        return min_link
    
    def vec(self):
        """Vectorize the current link based on its start node and end node.

        Raises:
            error.RegionError: Empty link.

        Returns:
            tuple: A tuple of coordinates that represents the vector.
        """
        if self.is_empty():
            raise error.RegionError("Empty link cannot be vectorized.")
        return (self.end.x - self.start.x, self.end.y - self.start.y)

    def reverse(self):
        """Reverse the direction of a link.

        Returns:
            Link: The link that has the opposite direction of the current link.
        """
        return Link(self.end, self.start)


class ValueDict(object):
    """Check whether certain values are in this dictionary. If not, it will add them automatically.

    Important note. You will find almost all dictionaries in this package using OrderedDict instead of the common dict.
    This is because we want to reach out to the same result when we use either Python 2 environment or Python 3 environment.
    In Python 2, the default dict object does not record the insertion order, whereas in Python 3 the default dict does.
    This will influence the iteration process when we cluster points or generate regions.
    In order to keep the consistent final result, we choose to use the OrderedDict object as out default dictionary.

    Attributes:
        value_dict(dict): The dictionary to store everything.
    """
    def __init__(self):
        """Construct an empty dictionary.
        """
        self.value_dict = OrderedDict()

    def find(self, key, value=None):
        """Add a key to the dictionary. 

        If the key exists in this dictionary, we return the value of this key.
        Else we return the insertion value of the key.
        Args:
            key (optional): The key of the dictionary.
            value (optional, optional): Defaults to None.

        Returns:
            Optional: The corresponding value.
        """
        key_str = key
        if key_str in self.value_dict:
            return self.value_dict[key_str]
        else:
            if value is None:
                val = key
            else:
                val = value
            self.value_dict[key_str] = val
            return val

    def is_in(self, key, value=None):
        """Check whether the key is in the dictionary.

        If not, add the value of the key to this dictionary.
        If the value is also None, we add the value as the key.

        Args:
            key (optional): The key of the dictionary.
            value (optional, optional): Defaults to None.

        Returns:
            bool: True if the key exists in the dictioanry.
        """
        key_str = key
        if key_str in self.value_dict:
            return True
        else:
            if value is None:
                val = key
            else:
                val = value
            self.value_dict[key_str] = val
            return False
        

class RegionGenerator(object):
    """The main region generator. 

    Attributes:
        links (list): A list that stores all links we need to use during region generation.
        node_dict (ValueDict): Keep track of all nodes being used.
        link_dict (dict): Keep track of all links being used.
    """
    def __init__(self, segs):
        """Initialize the region generator with segments.

        Args:
            segs (list): A list of mygeo.Segment objects.
        """
        self.links = []
        self.node_dict = ValueDict()
        # self.link_dict = {}
        self.link_dict = OrderedDict()
        for seg in segs:
            start = self.node_dict.find(seg.start.askey(), Node(seg.start.x, seg.start.y))
            end = self.node_dict.find(seg.end.askey(), Node(seg.end.x, seg.end.y))
            if start == end:     # When the start node and the end node of the segment are very close to each other.
                continue
            link = self.__get_link(start, end)
            if link is not None:
                self.links.append(link)
                self.links.append(link.reverse())

    def __get_link(self, start, end):
        """Generate a link using the start node and the end node.

        Args:
            start (node): The start node.
            end (node): The end node.

        Returns:
            Link: The directional link.
        """
        ret = None
        segstr = mygeo.Segment(start, end).askey()
        if segstr not in self.link_dict:
            self.link_dict[segstr] = 1
            ret = Link(start, end)
        return ret
        
    def run(self):
        """Generate the region.

        The basic idea for us to generate the region is to recursively search for the leftest link.
        When the outlink reachs a inlink, a ring is formed.

        Returns:
            List: A list of regions.
        """
        ret = []
        for link in self.links:
            if not link.used:
                reg = self.get_a_region(link) 
                if reg is not None:
                    ret.append(reg)
        return ret
    
    def get_a_region(self, link):
        """Get a region starting with the input link.

        Args:
            link (Link): The starting link.

        Raises:
            error.RegionError: The start node and the end node should be the same.
            error.RegionError: Too many rings are counter clockwise.

        Returns:
            mygeo.Region: The region formed by finding the leftest outlink recursively.
        """
        nodes = [link.start]
        next = link
        while not next.used:
            nodes.append(next.end)
            next.used = True
            next = next.leftest()
        if not nodes[0] == nodes[-1]:
            raise error.RegionError("The first node is not the same as the end node.")
        points = None
        holes = []
        node_dict = {}
        temp_holes = []
        i = 0
        while i < len(nodes):
            nd = nodes[i]
            ndstr = nd.askey()
            if ndstr in node_dict:
                start_idx = node_dict[ndstr]
                sub_nodes = nodes[start_idx: i]
                if len(sub_nodes) >= 3:
                    region = mygeo.Region(sub_nodes)
                    if region.area() > 0:
                        if mygeo.is_counter_clockwise(sub_nodes):
                            if points is None:
                                points = sub_nodes
                            else:
                                raise error.RegionError("Too many rings are counter clockwise")
                        else:
                            holes.append(sub_nodes)
                    else:
                        temp_holes.append(sub_nodes)
                nodes = nodes[:start_idx] + nodes[i:]
                i = start_idx
            else:
                node_dict[ndstr] = i
            i += 1
        if points is not None:
            return mygeo.Region(points, holes)
        else:
            return None

                    
def segments_to_points(segments):
    """Get all distinct points from a list of segments.

    Args:
        segments (list): A list of mygeo.Segment objects.

    Returns:
        List: A list of mygeo.Point object.
    """
    ret = []
    point_dict = ValueDict()
    for seg in segments:
        if not point_dict.is_in(seg.start.askey(), seg.start):
            ret.append(seg.start)
        if not point_dict.is_in(seg.end.askey(), seg.end):
            ret.append(seg.end)
    return ret


def segments_to_cluster_points(segments):
    """Get all distinct cluster points from a list of segments.

    Note that the difference between this function and the previous one is that
    it generates a list of Cluster.Point objects.

    Args:
        segments (list): A list of mygeo.Segment objects.

    Returns:
        list: A list of Cluster.point objects.
    """
    ret = []
    point_dict = ValueDict()
    for seg in segments:
        if not point_dict.is_in(seg.start.askey(), seg.start):
            ret.append(cluster.Point(seg.start.x, seg.start.y))
        if not point_dict.is_in(seg.end.askey(), seg.end):
            ret.append(cluster.Point(seg.end.x, seg.end.y))
    return ret


def regions_to_points(regions):
    """Get all distinct vluster points from a list of regions.

    Args:
        regions (list): A list of mygeo.Point objects.

    Returns:
        list: A list of mygeo.Point objects.
    """
    ret = []
    point_dict = ValueDict()
    for region in regions:
        for pt in region.points:
            if not point_dict.is_in(pt.askey(), pt):
                ret.append(pt)
        for hole in region.holes:
            for pt in hole:
                if not point_dict.is_in(pt.askey(), pt):
                    ret.append(pt)
    return ret


def regions_to_cluster_points(regions):
    """Get all distinct cluster points from a list of regions.

    Note that the difference between this function and the previous one is that
    it generates a list of Cluster.Point objects.

    Args:
        segments (list): A list of mygeo.Segment objects.

    Returns:
        list: A list of Cluster.point objects.
    """
    ret = []
    id = 0
    point_dict = ValueDict()
    for region in regions:
        id += 1
        for pt in region.points:
            if not point_dict.is_in(pt.askey(), pt):
                ret.append(cluster.Point(pt.x, pt.y, id))
        for hole in region.holes:
            for pt in hole:
                if not point_dict.is_in(pt.askey(), pt):
                    ret.append(cluster.Point(pt.x, pt.y, id))
    return ret


def clusters_to_pointmap(clusters):
    """Generate a point dictionary from a list of clusters.

    Args:
        clusters (list): A list of clusters.

    Returns:
        ValueDict: The point dictionary.
    """
    point_dict = ValueDict()
    for cluster in clusters:
        center = cluster.center()
        for ptstr in cluster.points:
            point_dict.is_in(ptstr, center)
    return point_dict


def __segment_2_newseg(segment, point_dict, seg_dict):
    """Revise the precision of the segment and remove if it is in the segment dictionary.

    Args:
        segment (mygeo.Segment): The target segment.
        point_dict (ValueDict): The point dictionary.
        seg_dict (ValueDict): The segment dictionary.

    Returns:
        mygeo.Segment: The precision-revised segment.
    """
    start = point_dict.find(segment.start.askey(), segment.start)
    end = point_dict.find(segment.end.askey(), segment.end)
    start.trunc()
    end.trunc()
    if start == end:
        return None
    newseg = mygeo.Segment(start, end)
    if seg_dict.is_in(newseg.askey(), newseg):
        return None
    return newseg


def simplify_by_pointmap(segments, point_dict):
    """ Get all precision-revised segments from a list of segments and a list of regions.

    Args:
        segments (list): A list of segments.
        point_dict (ValueDict): A point dictionary.

    Returns:
        list: A list of all precision-revised segments.
    """
    ret = []
    seg_dict = ValueDict()
    for seg in segments:
        newseg = __segment_2_newseg(seg, point_dict, seg_dict)
        if newseg is not None:
            ret.append(newseg)
    return ret


class RegionAttract(object):
    """Attract and merge regions.

    Attributes:
        grid_dict (dict): Grid segmentations.
        regions (list): The list of regions that needs to attract and merge.
        point_map (ValueDict): Store the revise all points that being used.
        width (int): The merging threshold.

    """
    def __init__(self, regions, width):
        """Initialization.

        Args:
            regions (list): A list of regions that needs to attract and merge.
            width (int): The merging threshold.
        """
        grid_dict = {}
        points = regions_to_points(regions)
        point_map = ValueDict()
        for pt in points:
            grid = pt.grid(width)
            item = [pt, None, width]
            if point_map.is_in(pt.askey(), item):
                continue
            for grid_x in range(grid[0] - 1, grid[0] + 2):
                for grid_y in range(grid[1] - 1, grid[1] + 2):
                    key_grid = (grid_x, grid_y)
                    if key_grid in grid_dict:
                        grid_dict[key_grid].append(item)
                    else:
                        grid_dict[key_grid] = [item]
        self.grid_dict = grid_dict
        self.regions = regions
        self.point_map = point_map
        self.width = width

    def __modify_ring(self, points):
        """Revise the precision of every point on the ring.

        Args:
            points (list): A list of points.

        Returns:
            list: A list of revised points.
        """
        new_points = []
        for pt in points:
            item = self.point_map.find(pt.askey(), None)
            if item is None or item[1] is None:
                new_points.append(pt)
            else:
                new_points.append(item[1])
        return new_points
    
    def __project(self, seg, point):
        """Find the location of the projection point on a segment.

        The project() function first finds a point on the LineString which has the shortest distance to the point.
        Then, it calculate how much it goes from the begining of the LineString.
        The interpolate() function returns the point that travels a certain distance from the begining of the line.

        Args:
            seg (mygeo.Segment): The given undirectional segment.
            point (mygeo.Point): The given point.

        Returns:
            mygeo.Point: The projection point.
        """
        l1 = seg.linestring()
        d1 = l1.project(point.point())
        pt = l1.interpolate(d1)
        return mygeo.Point(pt.x, pt.y)

    def run(self, segs):
        """Merge the existing regions to the segments.

        Args:
            segs (list): A list of segments.

        Returns:
            List: A list of new regions.
        """
        # Calculate the point projection on the segment and the distance.
        for seg in segs:
            for grid in seg.grids(self.width):
                if grid not in self.grid_dict:
                    continue
                for item in self.grid_dict[grid]:
                    (pt, spt, min_dist) = item
                    dist = mygeo.pt_2_seg_dist(pt, seg)
                    if dist < min_dist:
                        item[1] = self.__project(seg, pt)
                        item[2] = dist
        # Construct based on the point_map.
        ret = []
        for reg in self.regions:
            points = self.__modify_ring(reg.points)
            holes = []
            for hole in reg.holes:
                holes.append(self.__modify_ring(hole))
            newreg = None
            try:
                newreg = mygeo.Region(points, holes)
            except error.RegionError:
                error.debug("The region is not valid: %s" % (str(reg)))
            if newreg is not None and newreg.polygon().is_valid:
                ret.append(newreg)
            else:
                ret.append(reg)
        return ret




# -*- coding: utf-8 -*-
########################################################################
# 
# Copyright (c) 2019 Baidu.com, Inc. All Rights Reserved
# 
########################################################################

import sys
import gc
from collections import OrderedDict

from genregion.region import geometry as mygeo
from genregion.region import error

class SegSpliter(object):
    """The segments breaker.

    The main function of this object is to break segments into smaller pieces so that
    a segment only connects to others at its two end points. 

    Attributes:
        segs (list): A list of original segments that need to be processed.
        valids (list): A list of indicators that records the validity of each segment.
        grid_size (int): The size of the grid.
        grid_dict (dict): The dictionary that records grid information.
    """
    def __init__(self, segs, grid_size):
        """Initilize the spliter object.

        Args:
            segs (list): A list of original segments.
            grid_size (int): The size of the grid.
        """
        self.segs = segs
        self.valids = [ True for seg in segs ]
        self.grid_size = grid_size
        self.grid_dict = None
        
    def __proc_seg_by_seg(self, i1, i2):
        """Break two segments based on their intersection.
        
        After we find the intersection of two segments, we break both of them.
        Then we set the indicator for orginal segments to false, and add new pieces to the list.

        Args:
            i1 (int): The index of the first segment.
            i2 (int): The index of the second segment.

        Returns:
            list: New segments.
        """
        seg1 = self.segs[i1]
        seg2 = self.segs[i2]
        ret = []
        points = seg1.intersect(seg2)
        if len(points) > 0:
            for pt in points:
                pt.trunc()
            segs = self.__split_seg_by_points(seg1, points)
            if len(segs) > 1:
                ret.extend(segs)
                self.valids[i1] = False
            segs = self.__split_seg_by_points(seg2, points)
            if len(segs) > 1:
                ret.extend(segs)
                self.valids[i2] = False
        return ret

    def __split_seg_by_points(self, seg, points):
        """Break a segment into pieces based on a list of points.

        Args:
            seg (segment): The segment that needs to be broken.
            points (list): A list of points that should be on the segment.

        Returns:
            list: A list of broken segments.
        """
        pts = []
        line = seg.linestring() # Make the line directional
        for pt in points:
            if pt == seg.start or pt == seg.end:
                continue
            sppt = pt.point()
            dist = line.project(sppt) # Calculate the distance from the begining.
            pts.append((pt, dist))
        pts.sort(key=lambda x: x[1])
        pts.append((seg.end, 0))
        last_point = seg.start
        ret = []
        for pt, dist in pts:
            if last_point == pt:
                continue
            newseg = mygeo.Segment(last_point, pt)
            ret.append(newseg)
            last_point = pt
        return ret

    def __add_dict(self, indexes):
        """Add new segments to the grid dictionary.

        Args:
            indexes (list): A list of indexes indicating positions of segments in segs.
        """
        for i in indexes:
            seg = self.segs[i]
            for grid in seg.grids(self.grid_size):
                if grid in self.grid_dict:
                    self.grid_dict[grid].append(i)
                else:
                    self.grid_dict[grid] = [i]

    def __get_intersect_segs(self, i1):
        """Find all segments that intersect with a segment, and break them.

        Args:
            i1 (int): The index of the segment.

        Returns:
            list: A list of new segments.
        """
        ret = []
        seg = self.segs[i1]
        for grid in seg.grids(self.grid_size):
            if grid in self.grid_dict:
                for i2 in self.grid_dict[grid]:
                    seg2 = self.segs[i2]
                    if self.valids[i2]:
                        segs = self.__proc_seg_by_seg(i1, i2)
                        if len(segs) > 0:
                            ret.extend(segs)
                            if not self.valids[i1]:
                                return ret
        return ret

    def run(self):
        """Break all segments and filter out repeated ones.

        Returns:
            list: A list of new segments that only connect to others at their two end nodes.
        """
        # self.grid_dict = {}
        self.grid_dict = OrderedDict()
        i = 0
        while i < len(self.segs):
            if self.valids[i]:
                res_segs = self.__get_intersect_segs(i)
                if len(res_segs) > 0:
                    self.segs.extend(res_segs)
                    self.valids.extend([ True for rs in res_segs])
                if self.valids[i]:
                    self.__add_dict([i])
            i += 1
        ret = []
        seg_set = set([])
        for i in range(len(self.segs)):
            if self.valids[i]:
                seg = self.segs[i]
                seg_key = seg.askey()
                if seg_key not in seg_set:
                    seg_set.add(seg_key)
                    ret.append(seg)
        return ret



# -*- coding: utf-8 -*-
########################################################################
# 
# Copyright (c) 2021 Baidu.com, Inc. All Rights Reserved
# 
########################################################################
 
import sys
import datetime

def debug(info):
    """Print the debug information.

    Args:
        info (str): A sentence that indicates a message for debugging.
    """
    print(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), info)
    sys.stdout.flush()

class RegionError(Exception):
    """Region operation error.
    """
    def __init__(self, err_message):
        Exception.__init__(self, err_message)




# -*- coding: utf-8 -*-
########################################################################
# 
# Copyright (c) 2019 Baidu.com, Inc. All Rights Reserved
# 
########################################################################
 

import gc
import sys
from shapely.geometry import Point, LineString

from genregion.region import geometry as mygeo
from genregion.region import error
from genregion.generate.gen import cluster
from genregion.generate.gen import generator
from genregion.generate.gen import regionalg
from genregion.generate.gen import segspliter

def rect_cluster(points, width=40):
    """The hierarchical cluster algorithm.

    Args:
        points (list): A list of mygeo.Point objects ready to be clustered.
        width (int, optional): Clustering threshold. Defaults to 40.

    Returns:
        list: A list of clusters.
    """
    alg = cluster.HCAlgorithm(width)
    return alg.run(points)


def rect_classify(points, clusters, width=40):
    """Classify extra points to corresponding clusters.

    Args:
        points (list): A list of mygeo.Point objects.
        clusters (list): A list of hcclusters.
        width (int, optional): Threshold. Defaults to 40.

    Returns:
        list: A list of points which have not been classified.
    """
    alg = cluster.Classifier(clusters, width)
    return alg.run(points)


def region_filter(regions, grid_size=1024, thread_num=4):
    """Filter out subregions based from original regions.

    Args:
        regions (list): A list of regions that might contain subregions.
        grid_size (int, optional): Used to spilt regions to different threads. Defaults to 1024.
        thread_num (int, optional): The number of threads. Defaults to 4.

    Returns:
        list: A list of filtered regions.
    """
    ret = None
    rf = regionalg.RegionFilter(regions)
    if thread_num <= 1:
        ret = rf.run(grid_size)
    else:
        ret = rf.multi_run(grid_size, thread_num)
    return ret


def split_segs_by_segs(segs, grid_size=1024):
    """Split segments to prepare for region generation.

    We first calculate every intersection between each pair of segments.
    Then, we split those segments based on these intersections.
    The output of our function should be segments that connect to others only at its two end nodes.

    Args:
        segs (list): A list of segments.
        grid_size (int, optional): The size of the grid. Defaults to 1024.

    Returns:
        list: A list of segments.
    """
    ss = segspliter.SegSpliter(segs, grid_size)
    return ss.run()


def splitedsegs_2_regions(segs):
    """Generate regions based on processed segments.

    Args:
        segs (list): A list of segments that only connect to others at their two end points.

    Returns:
        list: A list of regions generated by input segments.
    """
    rg = generator.RegionGenerator(segs)
    return rg.run()


def merge_regions(regions, grid_size=1024, \
                  area_thres=10000, width_thres=20):
    """Merge raw regions with our regions. We temporarily stop this function.

    Args:
        regions (list): A list of regions.
        grid_size (int, optional): The size of the grid. Defaults to 1024.
        width_thres (int, optional): The width threshold to filter out narrow regions. Defaults to 20.

    Returns:
        list: A list of regions that incorporates with original regions.
    """
    raw_regions = None
    rm = regionalg.RegionMerger(regions, raw_regions)
    segs = rm.run(grid_size, area_thres, width_thres)
    rg = generator.RegionGenerator(segs)
    return rg.run()


def __cluster_points(segments, grid_size, clust_width):
    """Run the clustering algorithm on all points of all segments.

    Args:
        segments (list): A list of original segments.
        grid_size (int): The size of the grid.
        clust_width (int): The threshold of the cluster.

    Returns:
        list: A list of clusters.
    """
    if isinstance(segments, str):    # Read segments from the file.
        error.debug("Loading segments from the file: %s" % (segments))
        seg_points = generator.segments_to_cluster_points(mygeo.gen_segments(segments))
    else:
        error.debug("The number of segments: %d" % (len(segments)))
        seg_points = generator.segments_to_cluster_points(generate_segments(segments))
    error.debug("The number of different segment endpoints: %d" % (len(seg_points)))
    clusters = rect_cluster(seg_points, clust_width)
    del seg_points
    gc.collect()
    error.debug("The number of clusters: %d" % (len(clusters)))
    return clusters

def generate_segments(original_segments):
    """Convert a list of shapely linestrings, shapely point tuples, or coordinate pairs
        to a list of mygeo.Segment objects.

    Args:
        original_segments (list): A list of representations of segments.

    Raises:
        error.RegionError: Empty object.
        error.RegionError: Unsupported type of input.

    Returns:
        list: A list pf mygeo.Segment objects.
    """
    if original_segments is None:
        raise error.RegionError("Empty object")
    segs = []
    if isinstance(original_segments[0], tuple):
        for pt1, pt2 in original_segments:
            if isinstance(pt1, Point) and isinstance(pt2, Point):
                seg = mygeo.Segment(mygeo.Point(pt1.x, pt1.y), mygeo.Point(pt2.x, pt2.y))
            elif isinstance(pt1[0], float) or isinstance(pt1[0], int):
                seg = mygeo.Segment(mygeo.Point(pt1[0], pt1[1]), mygeo.Point(pt2[0], pt2[1]))
            else:
                raise error.RegionError("Unsupported type of input.")
            segs.append(seg)
    elif isinstance(original_segments[0], LineString):
        for linestr in original_segments:
            (x1, y1), (x2, y2) = list(linestr.coords)
            seg = mygeo.Segment(mygeo.Point(x1, y1), mygeo.Point(x2, y2))
            segs.append(seg)
    else:
        raise error.RegionError("Unsupported type of input.")
    return segs


def generate_regions(segments, grid_size=1024, \
                     area_thres=10000, width_thres=20, \
                     clust_width=25, point_precision=0):
    """Generate regions based on input segments.

    To represent original segments, we basically support 4 types of input:
        1. The file path that stores every segment of the road network:
            The file should only contain point coordinates at each line.
            Ex: 255834.51327326 3323376.71868603,260889.23516149 3321991.45674967
            Note that there is a comma between two points.
        2. A list of tuples of shapely points:
            Ex: [(Point_1, Point_2), (Point_3, Point_4), ...]
        3. A list of tuples of point coordinates:
            Ex: [((x1, y1), (x2, y2)), ((x3, y3), (x4, y4)), ...]
        4. A list of shapely LineString objects:
            Ex: [LineString_1, LineString_2]
    Important note: 
        All forms of segment representations must be a list of single segment.

    Args:
        segments (list): See details above.
        grid_size (int, optional): Use to build a grid dictionary for searching. Defaults to 1024.
        area_thres (int, optional): The minimum area of a generated region. Defaults to 10000.
        width_thres (int, optional): The minimum ratio of area/perimeter. Defaults to 20.
        clust_width (int, optional): The threshold that helps construct the cluster.
        point_precision (int, optional): The precision of the point object while processing.

    Returns:
        list: A list of generated regions.
    """
    mygeo.Point.set_precision(point_precision)
    clusters = __cluster_points(segments, grid_size, clust_width)
    pointmap = generator.clusters_to_pointmap(clusters)
    del clusters
    gc.collect()
    error.debug("The pointmap is finished.")
    if isinstance(segments, str):
        segs = generator.simplify_by_pointmap(mygeo.gen_segments(segments), pointmap)
    else:
        segs = generator.simplify_by_pointmap(generate_segments(segments), pointmap)
    del pointmap
    gc.collect()
    error.debug("The number of original segments: %d" % (len(segs)))
    segs = split_segs_by_segs(segs, grid_size)
    gc.collect()
    error.debug("The number of splitted segments.: %d" % (len(segs)))
    regs = splitedsegs_2_regions(segs)
    del segs
    gc.collect()
    error.debug("The number of the regions generated at the first step: %d" % (len(regs)))
    regs = merge_regions(regs, grid_size, area_thres, width_thres)
    gc.collect()
    error.debug("The number of regions after merging those small ones: %d" % (len(regs)))
    regs = region_filter(regs, grid_size)
    gc.collect()
    error.debug("The number of regions after filter out subregions: %d" % (len(regs)))
    return [region.polygon() for region in regs]




# -*- coding: utf-8 -*-
########################################################################
# 
# Copyright (c) 2019 Baidu.com, Inc. All Rights Reserved
# 
########################################################################
 
import math

from shapely import wkt
from shapely import geometry as spgeo
from shapely import ops as spops
from ordered_set import OrderedSet

from genregion.region import error

class Point(object):
    """ The point object.
    
    Attributes:
        x (float): The x-coordinate.
        y (float): The y-coordinate.
    
    """
    __slots__ = ("x", "y")
    # The precision of the coordinate.
    # If we are using projection coordinates like BD09MC or EPSG3587, we suggest Point.precision = 2.
    # If we are using the geodetic system like WGS84 and BD09, we suggest Point.precision = 5.
    precision = 2
    base = math.pow(10, precision)
    
    @classmethod
    def set_precision(cls, precision):
        """Set the precision of the point.

        Args:
            precision (int): The level of precision.
        """
        cls.precision = int(precision)
        cls.base = math.pow(10, cls.precision)

    def __init__(self, x, y):
        """Construct a point object from either the number or the string.

        Args:
            x (float): The x-coordinate.
            y (float): The y-coordinate.
        """
        self.x = float(x)
        self.y = float(y)

    def __eq__(self, pt):
        """Check whether both coordiates of two points are the same.

        Args:
            pt (Point): The other point.

        Returns:
            bool: True if they are the same.
        """
        return pt is not None and self.x == pt.x and self.y == pt.y

    def __str__(self):
        """Convert the point to a string format.

        Raises:
            error.RegionError: Empty object.

        Returns:
            str: The string format of the point.
        """
        if self.is_empty():
            raise error.RegionError("Cannot transfer an empty object to a string.")
        format = "%%.%df %%.%df" % (Point.precision, Point.precision)
        return format % (self.x, self.y)

    def trunc(self):
        """Truncate the point by precision.

        Raises:
            error.RegionError: Empty object.

        Returns:
            Point: Return the object itself.
        """
        if self.is_empty():
            raise error.RegionError("Cannot truncate an empty object.")
        self.x = int(self.x * Point.base + 0.5) * 1.0 / Point.base
        self.y = int(self.y * Point.base + 0.5) * 1.0 / Point.base
        return self

    def askey(self):
        """Convert the point object to a tuple of float as the dictionary key.

        Raises:
            error.RegionError: Empty object.

        Returns:
            tuple: A tuple of float represents the key.
        """
        if self.is_empty():
            raise error.RegionError("Empty object has no key.")
        return (self.x, self.y)

    def point(self):
        """Convert the point to the shapely point.

        Sometimes we need to use some functions in the shapely package such as interpolate() or project().

        Raises:
            error.RegionError: Empty object.

        Returns:
            spego.Point: The shapely format point.
        """
        if self.is_empty():
            raise error.RegionError("Cannot convert an empty object.")
        return spgeo.Point(self.x, self.y)

    def grid(self, grid_size=1024):
        """Find the grid which the point locates.

        To identify a grid, we use the integer coordinates for its position.

        Args:
            grid_size (int, optional): The width of the grid. Defaults to 1024.

        Raises:
            error.RegionError: Empty object.

        Returns:
            tuple: The horizontal and the vertical number which identify the order.
        """
        if self.is_empty():
            raise error.RegionError("Cannot find the grid for an empty object.")
        x_grid = int(self.x) // int(grid_size)
        y_grid = int(self.y) // int(grid_size)
        return (x_grid, y_grid)

    def destroy(self):
        """Destroy the point by setting its coordinates to None.
        """
        self.x = None
        self.y = None

    def is_empty(self):
        """Check whether the point object is empty.

        Returns:
            bool: True if not None.
        """
        return self.x is None


class Segment(object):
    """ The segment object, which is undirectional.

    Attributes:
        start (point): The start point.
        end (point): The end point.

    """
    __slots__ = ("start", "end")

    def __init__(self, start, end):
        """Initialize the segment using two points.

        Args:
            start (point): The start point.
            end (point): The end point.

        Raises:
            error.RegionError: Same point.
        """
        if start == end:
            raise error.RegionError("The start and the end point are the same.")
        self.start = start
        self.end = end

    def __eq__(self, seg):
        """Check whether two segments are the same.

        Args:
            seg (segment): The other segment.

        Returns:
            bool: True if the same.
        """
        if seg is not None and ((self.start == seg.start and self.end == seg.end) \
           or (self.end == seg.start and self.start == seg.end)):
            return True
        return False

    def __str__(self):
        """Convert a segment into a string.

        Raises:
            error.RegionError: Empty object.

        Returns:
            str: The string format like "x1 y1,x2 y2".
        """
        if self.is_empty():
            raise error.RegionError("Cannot convert an empty object.")
        start = str(self.start)
        end = str(self.end)
        if start < end:
            retstr = "%s,%s" % (start, end)
        else:
            retstr = "%s,%s" % (end, start)
        return retstr
    
    def askey(self):
        """Convert the segment object to a tuple of float as the dictionary key.

        Raises:
            error.RegionError: Empty object.

        Returns:
            tuple: A tuple of 4 float numbers represents the key.
        """
        if self.is_empty():
            raise error.RegionError("Empty object has no key.")
        if self.start.x < self.end.x \
                or (self.start.x == self.end.x and self.start.y < self.end.y):
            ret = (self.start.x, self.start.y, self.end.x, self.end.y)
        else:
            ret = (self.end.x, self.end.y, self.start.x, self.start.y)
        return ret

    def destroy(self):
        """Destroy the segment object by changing the start and end to None.
        """
        self.start = self.end = None

    def is_empty(self):
        """Check whether the segment is empty.

        Returns:
            bool: True if empty.
        """
        return self.start is None

    def length(self):
        """Calculate the length of the segment.

        Raises:
            error.RegionError: Empty object.

        Returns:
            float: The length of the segment.
        """
        if self.is_empty():
            raise error.RegionError("Empty object has no length.")
        return pt_2_pt_dist(self.start, self.end)

    def linestring(self):
        """Convert the segment into the directional shapely linestring.

        Raises:
            error.RegionError: Empty object.

        Returns:
            spgeo.LineString: The directional linestring.
        """
        if self.is_empty():
            raise error.RegionError("Cannot convert an empty segment.")
        points = []
        points.append((self.start.x, self.start.y))
        points.append((self.end.x, self.end.y))
        return spgeo.LineString(points)

    def __grids(self, x, y, grid_size, grid_set):
        """Return the grid that the given point locates.

        Note that there might be more than one grid being updated.
        If the point locates inside a grid, then we return one grid.
        If the point locates on the side of one grid, we return two grids.
        If the point locates on the vertex of one gird, we return four grids.

        Args:
            x (float): The x-coordinate of a point.
            y (float): The y-coordinate of a point.
            grid_size (int): The size of the grid.
            grid_set (set): The given set of the segments.
        """
        grid_x = int(x) // grid_size # changed
        grid_y = int(y) // grid_size # changed
        gs = [(grid_x, grid_y)]
        if grid_x * grid_size == x:
            gs.append((grid_x - 1, grid_y))
        if grid_y * grid_size == y:
            for i in range(len(gs)):
                gs.append((gs[i][0], grid_y - 1))
        for gd in gs:
            grid_set.add(gd)

    def grids(self, grid_size=1024):
        """Return all grids that a segment travels through.

        Args:
            grid_size (int, optional): The size of the grid. Defaults to 1024.

        Raises:
            error.RegionError: Empty object.

        Returns:
            list: A list of grid ids.
        """
        if self.is_empty():
            raise error.RegionError("Empty object is not in any grid.")
        grid_set = OrderedSet([])
        self.__grids(self.start.x, self.start.y, grid_size, grid_set)
        self.__grids(self.end.x, self.end.y, grid_size, grid_set)
        min_gx = int(min(self.start.x, self.end.x)) // grid_size # changed
        max_gx = int(max(self.start.x, self.end.x)) // grid_size # changed
        min_gy = int(min(self.start.y, self.end.y)) // grid_size # changed
        max_gy = int(max(self.start.y, self.end.y)) // grid_size # changed
        for grid_x in range(int(min_gx) + 1, int(max_gx) + 1):
            x = grid_x * grid_size
            y = self.start.y + (self.end.y - self.start.y) * (x - self.start.x) \
                    / (self.end.x - self.start.x)
            self.__grids(x, y, grid_size, grid_set)
        for grid_y in range(int(min_gy) + 1, int(max_gy) + 1):
            y = grid_y * grid_size
            x = self.start.x + (self.end.x - self.start.x) * (y - self.start.y) \
                    / (self.end.y - self.start.y)
            self.__grids(x, y, grid_size, grid_set)
        return list(grid_set)

    def intersect(self, seg):
        """Calculate the intersection with another segment.

        Note that there might be more or less than one intersection points of two segments.
        If two segments do not cross with each other, there is no intersection.
        If two segments overlap with each other, we return the starting and the ending overlapping point.
        If two segments merely intersect with each other, we return the intersection point.


        Args:
            seg (segment): The other segment.

        Returns:
            list: A list of intersection points.
        """
        thres = 0.001 / Point.base   # if t < thres, then we consider t = 0
        s01_x = self.end.x - self.start.x
        s01_y = self.end.y - self.start.y
        s23_x = seg.end.x - seg.start.x
        s23_y = seg.end.y - seg.start.y
        s20_x = self.start.x - seg.start.x
        s20_y = self.start.y - seg.start.y
        d01_len = math.sqrt(s01_x * s01_x + s01_y * s01_y)
        d23_len = math.sqrt(s23_x * s23_x + s23_y * s23_y)
        denom = s01_x * s23_y - s23_x * s01_y
        if denom == 0:  # Parallel or collinear
            d20_len = math.sqrt(s20_x * s20_x + s20_y * s20_y)
            cos201 = 0
            if d20_len > thres:
                cos201 = (s01_x * s20_x + s01_y * s20_y) / (d01_len * d20_len)
            sin201_2 = 1.0 - cos201 * cos201
            if sin201_2 < 0:
                sin201_2 = 0
            epson = d20_len * math.sqrt(sin201_2)
            if epson > 0.01 / Point.base:  # Parallel but not overlapped
                return []
            d1 = - d20_len * cos201
            if s01_x * s23_x + s01_y * s23_y > 0:
                d2 = d1 + d23_len
            else:
                d2 = d1 - d23_len
            d3 = d01_len
            if d2 < -thres and d1 < -thres or d2 > d3 + thres and d1 > d3 + thres:
                return []
            points = [(0, Point(self.start.x, self.start.y)), \
                      (d1, Point(seg.start.x, seg.start.y)), \
                      (d2, Point(seg.end.x, seg.end.y)), \
                      (d3, Point(self.end.x, self.end.y))]
            points.sort(key=lambda x: x[0])
            if points[1][1] == points[2][1]:
                return [points[1][1]]
            else:
                return [points[1][1], points[2][1]]
        s_numer = s01_x * s20_y - s01_y * s20_x
        t_numer = s23_x * s20_y - s23_y * s20_x
        s = s_numer / denom
        t = t_numer / denom
        if -s * d23_len > thres or (s - 1.0) * d23_len > thres:
            return []
        if -t * d01_len > thres or (t - 1.0) * d01_len > thres:
            return []
        if t >= 0 and t <= 1:
            if s < 0:
                return [Point(seg.start.x, seg.start.y)]
            if s > 1:
                return [Point(seg.end.x, seg.end.y)]
            else:
                return [Point(self.start.x + t * s01_x, self.start.y + t * s01_y)]
        if s >= 0 and s <= 1:
            if t < 0:
                return [Point(self.start.x, self.start.y)] 
            if t > 1:
                return [Point(self.end.x, self.end.y)]
        return []


class Region(object):
    """The region object.

    Attributes:
        points (list): A list of points that indicates the outer loop of polygon.
        holes (list): A list of lists of points that indicates holes.

    """
    __slots__ = ("points", "holes")

    def __filter_points(self, points):
        """Filter out the point which is the same as the previous one.

        Args:
            points (list): A list of points.

        Raises:
            error.RegionError: Less than 3 points.

        Returns:
            list: A list of filtered points.
        """
        if len(points) < 3:
            raise error.RegionError("Less than 3 points.")
        pts = []
        lastpt = None
        for pt in points:
            if not pt == lastpt:
                pts.append(pt)
                lastpt = pt
        if pts[0] == pts[-1]:
            pts.pop()
        return pts

    def __init_points(self, points, holes=None):
        """Initialize the points for constructing the polygon.

        The order of the points matters here.
        The order of points in the outer loop must be counterclockwise.
        The order of points in inner loops must be clockwise.
        Otherwise, we need to rearrange them in the correct way.

        Args:
            points (list): A list of points.
            holes (list, optional): A list of point lists representing holes. Defaults to None.

        Raises:
            error.RegionError: Less than 3 points.
            error.RegionError: Less than 3 points.
        """
        points = self.__filter_points(points)  # Filter out repeat points.
        if len(points) < 3:
            raise error.RegionError("A polygon cannot have less than 3 points.")
        if not is_counter_clockwise(points):    # The outer loop must be counterclockwise.
            self.points = points[::-1]
        else:
            self.points = points
        newholes = []
        if holes is not None and len(holes) > 0:
            for hole in holes:
                if len(hole) < 3:
                    raise error.RegionError("A polygon cannot have less than 3 points.")
                hole = self.__filter_points(hole)
                if is_counter_clockwise(hole):   # Inner loops must be clockwise.
                    newholes.append(hole[::-1])
                else:
                    newholes.append(hole)
        self.holes = newholes

    def __init_segments(self, segments):
        """Converting a list of segments into a list of points to construct the polygon.

        Note that we need to first convert those segments to a list of points and run the previous function.
        However, before doing that, we have to check a few conditions:
            1. The number of segments must be larger than 3.
            2. The starting point of each segment must be the same as the ending point of the previous one.
            3. The last segment and the first segment must be connected.
        After that, we will convert those segments into the point list for further polygon construction.

        Args:
            segments (list): A list of segments.

        Raises:
            error.RegionError: Less than 3 segments.
            error.RegionError: Consequtive segments are not connected.
            error.RegionError: The starting and the ending segment are not connected.
        """
        if len(segments) < 3:
            raise error.RegionError("The number of segments is less than 3.")
        points = [segments[0].start, segments[0].end]
        for seg in segments[1:]:
            if not seg.start == points[-1]:
                raise error.RegionError("Consequtive segments are not connected.")
            points.append(seg.end)
        if not points[0] == points[-1]:
            raise error.RegionError("The starting and the ending segment are not connected.")
        self.__init_points(points)
    
    def __init_polygon(self, polygon):
        """Convert a polygon to a list of points to construct the polygon.

        Args:
            polygon (spgeo.Polygon): The shapely format polygon.
        """
        points = []
        for pt in polygon.exterior.coords[:-1]:
            points.append(Point(pt[0], pt[1]))
        holes = []
        for inner in polygon.interiors:
            hole_points = []
            for pt in inner.coords[:-1]:
                hole_points.append(Point(pt[0], pt[1]))
            holes.append(hole_points)
        self.__init_points(points, holes)

    def __init_wkt(self, wktstr):
        """Convert a wkt string to a list of points to construct the polygon.

        Args:
            wktstr (str): The shapely wkt string format.

        Raises:
            error.RegionError: Not a shapely wkt string.
            error.RegionError: Loading failed.
        """
        try:
            geo = wkt.loads(wktstr)
            if geo.geom_type == "Polygon":
                self.__init_polygon(geo)
            else:
                raise error.RegionError("Not a polygon wkt string.")
        except Exception as e: 
            raise error.RegionError("Loading failed:" + e.message)

    def __init__(self, first, second=None):
        """Construct region points based on the given type of input.

        Args:
            first (list): Can be a list of points, segments, wkt strings or polygons 
            second (list, optional): Used for holes. Defaults to None.

        Raises:
            error.RegionError: Fail to initialize the region.
            error.RegionError: Fail to initialize the region.
        """
        if isinstance(first, list):
            if len(first) > 0:
                if isinstance(first[0], Point):
                    self.__init_points(first, second)
                elif isinstance(first[0], Segment):
                    self.__init_segments(first)
                else:
                    raise error.RegionError("Cannot initialize the region.")
        elif isinstance(first, spgeo.Polygon):
            self.__init_polygon(first)
        elif isinstance(first, str):
            self.__init_wkt(first)
        else:
            raise error.RegionError("Cannot initialize the region.")

    def __str__(self):
        """Convert the region to a wkt format string.

        Raises:
            error.RegionError: Empty object.

        Returns:
            str: The wkt format string.
        """
        if self.is_empty():
            raise error.RegionError("Cannot convert an empty object to a string.")
        ss = []
        for pt in self.points:
            ss.append(str(pt))
        ss.append(ss[0])
        wktstr = "POLYGON((" + ",".join(ss) + ")"
        if self.holes is not None:
            for hole in self.holes:
                ss = []
                for pt in hole:
                    ss.append(str(pt))
                ss.append(ss[0])
                wktstr += ",(" + ",".join(ss) + ")"
        wktstr += ")"
        return wktstr

    def assign(self, region):
        """Assign a new value to the region based on another region.

        Args:
            region (region): The target region.
        """
        self.points = region.points
        self.holes = region.holes
    
    def __ring_2_segments(self, points):
        """Generate a list of segments based on a list of points.

        Args:
            points (list): A list of points.

        Returns:
            list: A list of segments.
        """
        ret = []
        if points is None or len(points) < 2:
            return ret
        lastpt = points[-1]
        for pt in points:
            ret.append(Segment(lastpt, pt))
            lastpt = pt
        return ret

    def segments(self):
        """Convert the region into a list of segments.

        Raises:
            error.RegionError: Empty object.

        Returns:
            list: A list of segments.
        """
        if self.is_empty():
            raise error.RegionError("Cannot generate segments from an empty object.")
        ret = self.__ring_2_segments(self.points)
        for hole in self.holes:
            segs = self.__ring_2_segments(hole)
            ret.extend(segs)
        return ret

    def destroy(self):
        """Destroy a region by making everything empty.
        """
        self.points = self.holes = None

    def is_empty(self):
        """Check whether a region is empty.

        Returns:
            bool: True if empty.
        """
        return self.points is None

    def center(self):
        """ Return the center of the polygon.

        Note that this center might not be the centroid of the polygon.
        We basically use the representative point of the polygon.
        It is a cheaply computed point that is guaranteed to be within the geometric object.
        This function is mainly used to represent a polygon.

        Raises:
            error.RegionError: Empty object.

        Returns:
            point: The center of the object.
        """
        if self.is_empty():
            raise error.RegionError("Empty object does not have the center.")
        pt = self.polygon().representative_point()
        return Point(pt.x, pt.y)

    def mbr(self):
        """Calculate the MBR of the region.

        Raises:
            error.RegionError: Empty object.

        Returns:
            tuple: (minx, miny), (maxx, maxy)
        """
        if self.is_empty():
            raise error.RegionError("Cannot calculate the MBR for an empty region.")
        maxx = minx = self.points[0].x
        maxy = miny = self.points[0].y
        for pt in self.points[1:]:
            if pt.x < minx:
                minx = pt.x
            elif pt.x > maxx:
                maxx = pt.x
            if pt.y < miny:
                miny = pt.y
            elif pt.y > maxy:
                maxy = pt.y
        return (Point(minx, miny), Point(maxx, maxy))

    def grids(self, grid_size=1024):
        """Locate grids where the polygon lies in.

        Args:
            grid_size (int, optional): The size of the grid. Defaults to 1024.

        Raises:
            error.RegionError: Empty object.

        Returns:
            list: A list of grid ids (grid_x, grid_y).
        """
        if self.is_empty():
            raise error.RegionError("Cannot find the grid for an empty object.")
        bounds = self.mbr()
        start_grid = bounds[0].grid(grid_size)
        end_grid = bounds[1].grid(grid_size)
        ret = []
        for x_grid in range(start_grid[0], int(end_grid[0])):
            for y_grid in range(start_grid[1], int(end_grid[1])):
                ret.append((x_grid, y_grid))
        return ret

    def polygon(self):
        """Convert the region into shapely polygon.

        Raises:
            error.RegionError: Empty object.

        Returns:
            spgeo.Polygon: The shapely polygon.
        """
        if self.is_empty():
            raise error.RegionError("Cannot convert an empty region.")
        points = []
        for pt in self.points:
            points.append((pt.x, pt.y))
        holes = []
        if self.holes is not None:
            for hole in self.holes:
                hole_points = []
                for pt in hole:
                    hole_points.append((pt.x, pt.y))
                holes.append(hole_points)
        return spgeo.Polygon(points, holes)
        
    def __operate(self, op, region):
        """Run an operation on two polygons.

        Args:
            op (str): Can only be 'intersect', 'subtract' or union.
            region (region): The other region.

        Raises:
            error.RegionError: Empty object.
            error.RegionError: Unsupported operation type.

        Returns:
            list: A list of result regions. It only contains Polygon, Multipolygon or GeometryCollection.
        """
        if self.is_empty():
            raise error.RegionError("Cannot run a operation on an empty object.")
        ret = []
        polygon1 = self.polygon()
        polygon2 = region.polygon()
        if op == "intersect":
            inter = polygon1.intersection(polygon2)
        elif op == "subtract":
            inter = polygon1.difference(polygon2)
        elif op == "union":
            inter = polygon1.union(polygon2)
        else:
            raise error.RegionError("Unsupported operation type.")
        if inter.geom_type == "Polygon":
            ret.append(Region(inter))
        elif inter.geom_type == "MultiPolygon" or inter.geom_type == "GeometryCollection":
            for poly in inter:
                if poly.geom_type == "Polygon":
                    ret.append(Region(poly))
        return ret

    def intersect(self, region):
        """The intersection of two regions.

        Args:
            region (region): Another region.

        Returns:
            list: A list of regions.
        """
        return self.__operate("intersect", region)

    def subtract(self, region):
        """The difference of two regions.

        Args:
            region (region): Another region.

        Returns:
            list: A list of regions.
        """
        return self.__operate("subtract", region)

    def union(self, region):
        """The union of two regions.

        Args:
            region (region): Another region.

        Returns:
            list: A list of regions.
        """
        return self.__operate("union", region)

    def area(self):
        """Calculate the area of this region.

        Raises:
            error.RegionError: Empty object.

        Returns:
            float: The area value.
        """
        if self.is_empty():
            raise error.RegionError("Empty object has no area.")
        ret = []
        polygon = self.polygon()
        return polygon.area

    def length(self):
        """Calculate the perimeter of the region.

        Raises:
            error.RegionError: Empty object.

        Returns:
            float: The perimeter value.
        """
        if self.is_empty():
            raise error.RegionError("Empty object has no perimeter.")
        polygon = self.polygon()
        return polygon.length

    def gridize(self, grid_size=1024):
        """Cut a region into several pieces according to the grid.

        Args:
            grid_size (int, optional): Defaults to 1024.

        Raises:
            error.RegionError: Empty object.

        Returns:
            list: A list of result polygons.
        """
        if self.is_empty():
            raise error.RegionError("Empty object cannot be cut.")
        points = []
        bounds = self.mbr()
        start_grid = bounds[0].grid(grid_size)
        end_grid = bounds[1].grid(grid_size)
        minx = (start_grid[0] - 1) * int(grid_size)
        maxx = (end_grid[0] + 2) * int(grid_size)
        miny = (start_grid[1] - 1) * int(grid_size) 
        maxy = (end_grid[1] + 2) * int(grid_size)
        use_min = True
        for x_grid in range(int(start_grid[0]) + 1, int(end_grid[0]) + 1):
            x = x_grid * int(grid_size)
            if use_min:
                points.append((x, miny))
                points.append((x, maxy))
            else:
                points.append((x, maxy))
                points.append((x, miny))
            use_min = not use_min
        if use_min:
            y_list = range(start_grid[1] + 1, end_grid[1] + 1)
        else:
            y_list = range(end_grid[1], start_grid[1], -1)
        use_max = True
        for y_grid in y_list:
            y = y_grid * int(grid_size)
            if use_max:
                points.append((maxx, y))
                points.append((minx, y))
            else:
                points.append((minx, y))
                points.append((maxx, y))
            use_max = not use_max
        ret = []
        if len(points) > 0:
            line = spgeo.LineString(points)
            polygons = spops.split(self.polygon(), line)
            for poly in polygons:
                if poly.geom_type == "Polygon":
                    ret.append(Region(poly))
        else:
            ret.append(self)
        return ret

    def contains(self, point):
        """Check whether a point is in this region.

        Args:
            point (point): The point object.

        Raises:
            error.RegionError: Empty object.

        Returns:
            bool: True if the point is in this polygon.
        """
        if self.is_empty():
            raise error.RegionError("Empty object cannot have a point.")
        return self.polygon().covers(point.point())


def is_counter_clockwise(points):
    """Check whether a list of points are counterclockwise.

    Args:
        points (list): A list of points.

    Returns:
        bool: True if the arrangement of those points is counterclockwise.
    """
    pts = []
    for pt in points:
        pts.append((pt.x, pt.y))
    ring = spgeo.LinearRing(pts)
    return ring.is_ccw


def pt_2_pt_dist(pt1, pt2):
    """Calculate the distance between two points.

    Args:
        pt1 (point): The point object.
        pt2 (point): The point object.

    Returns:
        float: The distance of those two points.
    """
    return pt1.point().distance(pt2.point())


def pt_2_seg_dist(pt, seg):
    """Calculate the distance between a point and a segment.

    Args:
        pt (point): The point object.
        seg (segment): The segment object.

    Returns:
        float: The distance.
    """
    return pt.point().distance(seg.linestring())


def pt_2_reg_dist(pt, reg):
    """Calculate the distance of a point to a region.

    If the point is on the edge of the region, we return 0.
    If the point locates inside the region, we return a negative distance value.
    If the point locates outside the region, we return a positive distance value.

    Args:
        pt (point): The point object.
        reg (region): The region object.

    Returns:
        float: Distance that can be negative, positive or 0.
    """
    polygon = reg.polygon()
    point = pt.point()
    is_cover = polygon.covers(point)
    ring = spgeo.LinearRing(polygon.exterior.coords)
    min_dist = ring.distance(point)
    for hole in polygon.interiors:
        ring = spgeo.LinearRing(hole.coords)
        dist = ring.distance(point)
        if dist < min_dist:
            min_dist = dist
    if is_cover:
        return -min_dist
    else:
        return min_dist


def line_simp(points, threshold):
    """Perform Douglas simplification on a list of points.

    Args:
        points (list): A list of point objects that need to be simplified.
        threshold (float): Tolerance of the algorithm. 

    Returns:
        list: A list of simplified points.
    """
    pts = []
    for pt in points:
        pts.append((pt.x, pt.y))
    line = spgeo.LineString(pts)
    newline = line.simplify(threshold, preserve_topology=False)
    newpoints = []
    for npt in newline.coords[:]:
        newpoints.append(Point(npt[0], npt[1]))
    return newpoints


def dump_segments(segments, seg_file):
    """Save all segments to an local file.

    Args:
        segments (list): A list of segments.
        seg_file (str): The file path.
    """
    with open(seg_file, "w") as fo:
        for seg in segments:
            fo.write("%s\n" % (str(seg)))


def gen_segments(seg_file):
    """Segment generator.

    Args:
        seg_file (str): The file path to read those segments.

    Yields:
        Segment: Segment objects.
    """
    with open(seg_file, "rb") as fi:
        line = fi.readline()
        while line.decode() != "":
            p1, p2 = line.decode().strip().split(",")
            x1, y1 = p1.split(" ")
            x2, y2 = p2.split(" ")
            yield Segment(Point(float(x1), float(y1)), Point(float(x2), float(y2)))
            line = fi.readline()


def load_segments(seg_file):
    """Load a list of segments from a file.

    Args:
        seg_file (str): The file path to load those segments.

    Returns:
        list: A list of segments.
    """
    segs = []
    for seg in gen_segments(seg_file):
        segs.append(seg)
    return segs


def dump_regions(regions, reg_file):
    """Save all regions to a local file.

    Args:
        regions (list): A list of regions.
        reg_file (str): The file path to save all regions.
    """
    with open(reg_file, "w") as fo:
        for reg in regions:
            fo.write("%s\n" % (str(reg)))


def gen_regions(reg_file):
    """Region generator.

    Args:
        reg_file (str): The file path that stores those regions.

    Yields:
        Region: Region object.
    """
    with open(reg_file, "rb") as fi:
        line = fi.readline()
        while line != "":
            yield Region(line.strip())
            line = fi.readline()


def load_regions(reg_file):
    """Load all regions from a file.

    Args:
        reg_file (str): The file path.

    Returns:
        list: A list of regions.
    """
    regions = []
    for reg in gen_regions(reg_file):
        regions.append(reg)
    return regions




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
################################################################################
#
# Copyright (c) 2019 Baidu.com, Inc. All Rights Reserved
#
################################################################################

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="genregion",                     
    version="0.0.1",                        
    author="Yanyan Li & Ming Zhang",                  
    description="Region generation using urban road netwok",
    long_description=long_description,     
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),    
    py_modules=["genregion"],           
    install_requires=['ordered_set', "shapely"]
)




# -*- coding: utf-8 -*-
########################################################################
# 
# Copyright (c) 2019 Baidu.com, Inc. All Rights Reserved
# 
########################################################################
 
from .gen.cluster import Cluster
from .gen.algorithm import generate_regions
from .gen.algorithm import rect_cluster
from .gen.algorithm import rect_classify
from .gen.algorithm import region_filter
from .gen.algorithm import split_segs_by_segs
from .gen.algorithm import splitedsegs_2_regions
from .gen.algorithm import merge_regions
from .gen.regionalg import RegionFinder




# -*- coding: utf-8 -*-
########################################################################
# 
# Copyright (c) 2019 Baidu.com, Inc. All Rights Reserved
# 
########################################################################

import sys
from collections import OrderedDict
from ordered_set import OrderedSet

from genregion.region import error
from genregion.region import geometry as mygeo

class Point(mygeo.Point):
    """Point object for clustering.
    
    Attributes:
        x (float): The x-coordinate.
        y (float): The y-coordinate.
        id: The point id.
    """
    #__slots__ = ("id", )

    def __init__(self, x, y, id=None):
        mygeo.Point.__init__(self, x, y)
        self.id = id

class Cluster(object):
    """object that records clustering information.

    Attributes:
        minx (float): The smallest x-coordinate among all points in this cluster.
        maxx (float): The largest x-coordinate among all points in this cluster.
        miny (float): The smallest y-coordinate among all points in this cluster.
        maxy (float): The largest x-coordinate among all points in this cluster.
        points (set): Contains all points in this cluster.
        ids (set): Contains all point ids in this cluster.
    """
    #__slots__ = ("minx", "maxx", "miny", "maxy", "points", "ids")

    def __init__(self):
        """A cluster object contains information about MBR and points.
        """
        self.minx = None
        self.maxx = None
        self.miny = None
        self.maxy = None
        # self.points = set([])
        # self.ids = set([])
        self.points = OrderedSet([])
        self.ids = OrderedSet([])

    def count(self):
        """Count the number of the points in this cluster.

        Returns:
            int: The number of the points.
        """
        return len(self.points)

    def destroy(self):
        """Destroy the cluster.
        """
        self.points.clear()
        self.ids.clear()
        self.minx = self.maxx = self.miny = self.maxy = None

    def is_empty(self):
        """Check whether the cluster is empty.

        Returns:
            bool: True if empty.
        """
        return self.minx is None

    def add_point(self, point, is_vip=True):
        """Add a new point to this cluster.

        Args:
            point (Point): The new point.
            is_vip (bool, optional): It is used to calculate the center of the cluster. Defaults to True.
        """
        self.points.add(point.askey())
        if is_vip:
            if self.minx is None or point.x < self.minx:
                self.minx = point.x
            if self.maxx is None or point.x > self.maxx:
                self.maxx = point.x
            if self.miny is None or point.y < self.miny:
                self.miny = point.y
            if self.maxy is None or point.y > self.maxy:
                self.maxy = point.y
        if point.id is not None:
            self.ids.add(point.id)

    def merge(self, cl):
        """Merge a new cluster to the current one.

        Args:
            cl (Cluster): The other cluster.

        Raises:
            error.RegionError: Id conflict.
        """
        if cl.is_empty():
            return
        if not self.can_merge(cl):
            raise error.RegionError("Cluster id conflict.")
        if self.is_empty():
            self.minx, self.maxx, self.miny, self.maxy = cl.minx, cl.maxx, cl.miny, cl.maxy
        else:
            self.minx = min(self.minx, cl.minx)
            self.maxx = max(self.maxx, cl.maxx)
            self.miny = min(self.miny, cl.miny)
            self.maxy = max(self.maxy, cl.maxy)
        for pt in cl.points:
            self.points.add(pt)
        for id in cl.ids:
            self.ids.add(id)

    def can_merge(self, cl):
        """Check whether one cluster can be merged with the current one. 

        Args:
            cl (Cluster): The other cluster.

        Returns:
            bool: True if possible.
        """
        for id in cl.ids:
            if id in self.ids:
                return False
        return True

    def in_cluster(self, point):
        """Check whether a point is in the current cluster.

        Args:
            point (Point): The checking point.

        Returns:
            bool: True if contained.
        """
        return (point.askey() in self.points)

    def center(self):
        """Get the MBR center of all vip points as the cluster center.

        Raises:
            error.RegionError: Empty cluster.

        Returns:
            mygeo.Point: A point contains the 2D coordinates.
        """
        if self.is_empty():
            raise error.RegionError("Empty cluster has no center.")
        return mygeo.Point((self.minx + self.maxx) / 2, (self.miny + self.maxy) / 2)

    def width_x(self):
        """Calculate the horizontal width of the cluster.

        Raises:
            error.RegionError: Empty cluster.

        Returns:
           float: Width. None if empty.
        """
        if self.is_empty():
            raise error.RegionError("Empty cluster has no width_x.")
        return self.maxx - self.minx

    def width_y(self):
        """Calculate the vertical width of the cluster.

        Raises:
            error.RegionError: Empty cluster.

        Returns:
            float: Width. None if empty.
        """
        if self.is_empty():
            raise error.RegionError("Empty cluster has no width_y.")
        return self.maxy - self.miny


def is_friend_dist(dist, width):
    """Check whether the width is larger than the threshold.

    Args:
        dist (tuple): (width_x, width_y)
        width (int): The threshold.

    Returns:
        bool: True if possible.
    """
    return dist[0] <= width and dist[1] <= width


class HCCluster(Cluster):
    """Hierarchical clustering object.

    Attributes:
    minx (float): The smallest x-coordinate among all points in this cluster.
    maxx (float): The largest x-coordinate among all points in this cluster.
    miny (float): The smallest y-coordinate among all points in this cluster.
    maxy (float): The largest x-coordinate among all points in this cluster.
    points (set): Contains all points in this cluster.
    ids (set): Contains all point ids in this cluster.
    friends (list): Contains all friend clusters.

    """
    def __init__(self):
        """Initialization of a hierarchical cluster.

        In this object, we add a new variable called 'friend'. The friends of a cluster are its neighbors.
        Two clusters are friends if they merge into one cluster, of which each side is less than the threshold.

        """
        Cluster.__init__(self)
        # (cluster, dist)
        # The best friend is self.friends[0]
        self.friends = []

    def collect_best_friend(self):
        """Move the best friends to the begining of the friends list. 

        Note that there might be more than one best friends for each hccluster.

        """
        min_weight = sys.float_info.max
        for friend in self.friends:
            weight = friend[1]
            if weight < min_weight:
                min_weight = weight
        swap_count = 0
        i = 0
        while i < len(self.friends):
            if self.friends[i][1] == min_weight:
                self.friends[swap_count], self.friends[i] = self.friends[i], self.friends[swap_count]
                swap_count += 1
            i += 1

    def del_friends(self, del_frds):
        """Remove a list of friends from the current friends list.

        If the first item of the friend list get removed, we need to elect best friends again.

        Args:
            del_frds (list of friends): Unwanted friends.

        Returns:
            bool: True if best friends are reelected.
        """
        i = 0
        first_del = False
        while i < len(self.friends): 
            hc, weight = self.friends[i]
            if hc in del_frds:
                self.friends.pop(i)
                i -= 1
                if i == 0:
                    first_del = True
            i += 1
        if first_del and len(self.friends) > 1:
            self.collect_best_friend()
        return first_del

    def bfs_merge(self, width):
        """Check the condition and merge the cluster with its best friends.

        Two clusters will be merged if they are best friends of each other.

        Args:
            width (int): Merging threshold.

        Returns:
            bool: True if the merging process happens.
        """
        best_frds = self.best_friends()
        for hc, weight in best_frds:
            frds = hc.best_friends()
            for (hc2, weight) in frds:
                if self == hc2:
                    self.merge(hc, width)
                    return True
        return False

    def merge(self, hc, width):
        """Merge two hccluster objects together.
        
        Merge the current cluster with the other one. 
        Update the friends list of the combined cluster.

        Args:
            hc (hccluster): The merging hccluster.
            width (int): The merging threshold.
        """
        Cluster.merge(self, hc)
        new_friends = OrderedSet([])
        for fhc, weight in self.friends:
            new_friends.add(fhc)
        for fhc, weight in hc.friends:
            new_friends.add(fhc)
        hc.destroy()
        self.friends = []
        for fhc in new_friends:
            if fhc == self or fhc == hc:
                continue
            fhc.del_friends([self, hc])
            dist = self.distance(fhc, width)
            if is_friend_dist(dist, width) and self.can_merge(fhc):
                self.add_friend(fhc, dist)
                fhc.add_friend(self, dist)

    def best_friends(self):
        """Find all best friends of the current cluster.

        Returns:
            List: Best friends.
        """
        ret = []
        if len(self.friends) > 0:
            min_weight = self.friends[0][1]
            for friend in self.friends:
                if min_weight == friend[1]:
                    ret.append(friend)
        return ret

    def distance(self, hc, width):
        """Calculate the distance between two HCCluster objects.

        Definition of the distance between two clusters:
            If the widths of cluster A are (xa, ya), and the widths of cluster B are (xb, yb), 
            then the widths of the merged cluster C are (xc, yc), and they should both less than the width.
            So, the distance should be defined as:
                if xa + ya > xb + yb: 
                    delta_x = xc - xa
                    delta_y = yc - ya 
                else
                    delta_x = xc - xb
                    delta_y = yc - yb 

        Args:
            hc (hccluster): The merging hccluster.
            width (int): The merging threshold.

        Raises:
            error.RegionError: Empty object.

        Returns:
            tuple: (width_x, width_y)
        """
        if hc.is_empty() or self.is_empty():
            raise error.RegionError("Empty object has no distance.")
        minx = min(self.minx, hc.minx)
        maxx = max(self.maxx, hc.maxx)
        miny = min(self.miny, hc.miny)
        maxy = max(self.maxy, hc.maxy)
        width_x = maxx - minx
        width_y = maxy - miny
        if width_x <= width and width_y <= width:
            if self.width_x() + self.width_y() > hc.width_x() + hc.width_y():
                return (width_x - self.width_x(), width_y - self.width_y())
            else:
                return (width_x - hc.width_x(), width_y - hc.width_y()) 
        return (width + 1, width + 1)

    def add_friend(self, hc, dist):
        """Add a new cluster to the friends list.

        Args:
            hc (hccluster): The new cluster.
            dist (Tuple): Distance.
        """
        (dist_x, dist_y) = dist
        weight = dist_x + dist_y
        if len(self.friends) > 0 and weight <= self.friends[0][1]:
            self.friends.insert(0, (hc, weight))
        else:
            self.friends.append((hc, weight))

    def display(self):
        """Print detailed information about the current cluster.
        """
        print("hc cluster: ")
        for pt in self.points:
            print(pt, ",")
        for fhc, weight in self.friends:
            print("friend weight: ", weight, " : ")
            for pt in fhc.points:
                print(pt, ",")
        print("mbr: (%f, %f), (%f, %f)" % (self.minx, self.miny, self.maxx, self.maxy))


class HCAlgorithm(object):
    """The main object to run the hierarchical clustering algorithm.

    Attributes:
        width (int): Clustering threshold.
        hcclusters (list): A list of hcclusters that perform the clustering algorithm.
        grids (dict): Record all grids that being used for this algorithm.

    """
    def __init__(self, width):
        """The implementation of the clustering algorithm.

        Args:
            width (int): Clustering threshold.
        """
        self.width = width
        # Intermediate results of the clustering algorithm.
        self.hcclusters = []
        # dict of (grid_id, HCCluster list)
        # self.grids = {}
        self.grids = OrderedDict()

    # Build for debugging
    def output_points(self, points, path="seg_points.txt"):
        """Output all points of a given list of points.

        Args:
            points (list): segment points
            path (str, optional): . Defaults to "seg_points.txt".

        """
        with open(path, "w") as fw:
            for pt in points:
                fw.write(str(pt.askey())+'\n')

    # Build for debugging
    def output_grids(self, path="grid_id.txt"):
        """Output all grid ids.

        Args:
            path (str, optional): Defaults to "grid_id.txt".
        """
        with open(path, "w") as fw:
            for grid_id in self.grids:
                fw.write(str(grid_id) + "\n")

    # Build for debugging
    def output_clusters(self, ret, path="cluster.txt"):
        """Output all clusters.

        Args:
            ret (list): A list of hcc clusters.
            path (str, optional): Defaults to "cluster.txt".
        """
        with open(path, "w") as fw:
            for hc in ret:
                fw.write("%f, %f, %f, %f, %d\n" % (hc.minx, hc.maxx, hc.miny, hc.maxy, len(hc.points)))

    def run(self, points):
        """The implementation of the hierarchical clustering algorithm.
        Args:
            points (list): A list of mygeo.Point ready to be clustered.

        Returns:
            list: A list of clusters.
        """
        # Make each point as the initial cluster.
        self.build_clusters(points)
        # Construct the friendship between each pair of clusters.
        self.build_friends()
        # Deep first merge.
        merge_count = 1
        while merge_count > 0:
            merge_count = 0
            for hc in self.hcclusters:
                if not hc.is_empty():
                    if hc.bfs_merge(self.width):
                        merge_count += 1
        # Append each cluster to the list.
        ret = []
        for hc in self.hcclusters:
            if not hc.is_empty():
                ret.append(hc)
        # error.debug("Number of grids: %d" % len(self.grids)) 
        # wrting grid_id for debugging
        # self.output_grids()
        # self.output_clusters(ret)
        # self.output_points(points)
        return ret
        
    def build_clusters(self, points):
        """Construct initial clusters based on each point and assign them to each grid.

        Args:
            points (list): A list of mygeo.Point.
        """
        for pt in points:
            hc = HCCluster()
            hc.add_point(pt)
            self.hcclusters.append(hc)
            grid = pt.grid(self.width)    # grid is a tuple (grid_x, grid_y) 
            if grid in self.grids:
                self.grids[grid].append(hc)
            else:
                self.grids[grid] = [hc]
    
    def build_friends(self):
        """Construct the friendship among adjacent grids.
        """
        for grid in self.grids:
            grid_x, grid_y = grid
            self.build_friend_one_grid(grid)
            self.build_friend_two_grid(grid, (grid_x - 1, grid_y + 1))
            self.build_friend_two_grid(grid, (grid_x + 1, grid_y))
            self.build_friend_two_grid(grid, (grid_x, grid_y + 1))
            self.build_friend_two_grid(grid, (grid_x + 1, grid_y + 1))

    def build_friend_one_grid(self, grid):
        """Two clusters in the same grid must be friends.

        Args:
            grid (tuple): (grid_x, grid_y). Horizontal and vertical ids that identify a grid.
        """
        hcs = self.grids[grid]
        count = len(hcs)
        for i in range(count - 1):
            hc1 = hcs[i]
            for j in range(i + 1, count):
                hc2 = hcs[j]
                dist = hc1.distance(hc2, self.width)
                if is_friend_dist(dist, self.width) and hc1.can_merge(hc2):
                    hc1.add_friend(hc2, dist)
                    hc2.add_friend(hc1, dist)

    def build_friend_two_grid(self, grid1, grid2):
        """Build friendship between all clusters in each grid.

        Args:
            grid1 (tuple): (grid_x, grid_y).
            grid2 (tuple): (grid_x, grid_y).
        """
        if grid1 not in self.grids or grid2 not in self.grids:
            return
        hcs1 = self.grids[grid1]
        hcs2 = self.grids[grid2]
        for hc1 in hcs1:
            for hc2 in hcs2:
                dist = hc1.distance(hc2, self.width)
                if is_friend_dist(dist, self.width) and hc1.can_merge(hc2):
                    hc1.add_friend(hc2, dist)
                    hc2.add_friend(hc1, dist)
        

class Classifier(object):
    """ Put the point into a cluster according to the distance.

    Attributes:
        width (int): The clustering threshold.
        grids (int): All grids that being used.

    """
    def __init__(self, clusters, width):
        """Initialization.

        Args:
            clusters (list): A list of clusters.
            width (int): Threshold.
        """
        self.width = width
        self.grids = {}
        for cluster in clusters:
            center = cluster.center()
            lb = mygeo.Point(center.x - width / 2.0, center.y - width / 2.0)
            rt = mygeo.Point(center.x + width / 2.0, center.y + width / 2.0)
            lb_grid = lb.grid(width)
            rt_grid = rt.grid(width)
            for x in range(lb_grid[0], rt_grid[0] + 1):
                for y in range(lb_grid[1], rt_grid[1] + 1):
                    grid = (x, y)
                    if grid in self.grids:
                        self.grids[grid].append(cluster)
                    else:
                        self.grids[grid] = [cluster]
    
    def run(self, points):
        """Implementation of the distribution process.

        Args:
            points (list): A list of mygeo.Point objects.

        Returns:
            list: A list of mygeo.Point objects which cannot be classified.
        """
        left_points = []
        for pt in points:
            if not self.classify_point(pt):
                left_points.append(pt)
        return left_points

    def classify_point(self, point):
        """Classify the point to a cluster. 

        Args:
            point (mygeo.Point): The target point.

        Returns:
            bool: True if the point can be classified.
        """
        grid = point.grid(self.width)
        ret = False
        if grid in self.grids:
            clusters = self.grids[grid]
            max_count = 0
            min_cluster = None
            for cluster in clusters:
                center = cluster.center()
                dist_x = abs(center.x - point.x)
                dist_y = abs(center.y - point.y)
                if dist_x <= self.width / 2.0 and dist_y <= self.width / 2.0 \
                        and point.id not in cluster.ids:
                    count = cluster.count()
                    if count > max_count:
                        min_cluster = cluster
                        max_count = count
            if min_cluster is not None:
                min_cluster.add_point(point, False)
                ret = True
        return ret


